from typing import NamedTuple
from math import radians, cos, sin
import numpy as np
import xarray as xr
import sasktran as sk
from osirisl1services.spatialrotation import Vector, UnitQuaternion, TaitBryanZYX, ECI
from osirisl1services.readlevel1 import Orbit, open_level1_attitude, open_level1_spectrograph
from astropy.time import Time
from astropy import units as u
from astropy.coordinates import EarthLocation, AltAz, get_sun
import warnings
import logging


class PixelParams(NamedTuple):
    dy: float
    dz: float
    y0: float
    z0: float


# Spectrograph alignment data from OsirisConfigurationDatabase/flightduration/osoptics.ini
# The rotation is expressed as a euler angles "zyx" in degrees.
# 2005-01-03 NDL Initialize with the Jupiter alignment measurements.
SPECTROGRAPH_ROTATION_TB_ZYX = TaitBryanZYX(roll=radians(0.0337), pitch=radians(-0.0903), yaw=radians(-0.15810304))
SPECTROGRAPH_ROTATION_CF = SPECTROGRAPH_ROTATION_TB_ZYX.to_quaternion()


# InfraRed alignment data from OsirisConfigurationDatabase/flightduration/iroptics.ini
# The rotation is expressed as a euler angles "zyx" in degrees.
IR_ROTATION_TB_ZYX = TaitBryanZYX(roll=radians(0.0), pitch=radians(-0.06167), yaw=radians(0.0))
IR_ROTATION_CF = IR_ROTATION_TB_ZYX.to_quaternion()
MINUTES_TO_RADIANS = radians(1.0 / 60.0)
# IR1 Optics
IR1_PIXEL_PARAMS = PixelParams(dy=-1.1931386 * MINUTES_TO_RADIANS, dz=0.0, y0=105.104, z0=0.0)
# IR2 Optics
IR2_PIXEL_PARAMS = PixelParams(dy=-1.2046809 * MINUTES_TO_RADIANS, dz=0.0, y0=109.289, z0=0.0)
# IR3 Optics
IR3_PIXEL_PARAMS = PixelParams(dy=-1.2059574 * MINUTES_TO_RADIANS, dz=0.0, y0=109.492, z0=0.0)


@xr.register_dataset_accessor('l1')
class Level1Services(object):
    def __init__(self, xarray_ds):
        self._ds = xarray_ds
        self._attitude = None
        self._position_eci = None
        self._position_ecef = None
        self._look_eci = None
        self._look_ecef = None
        self._latitude = None
        self._longitude = None
        self._altitude = None
        self._sza = None

        self._exposuremjd = self._ds.mjd + 0.5 * self._ds.exposureTime / 86400.0

        if 'orbit' in self._ds:
            orbit = self._ds.orbit
        else:
            orbit = self._ds.scanno // 1000
        attitude1 = open_level1_attitude(Orbit(orbit), load_pointing=True)

        if self._ds.mjd.min() >= attitude1.mjd.min():
            self._attitude = attitude1
        else:
            attitude1 = attitude1.drop('ScanNumber').drop(
                [var for var in attitude1.data_vars if 'mjd' not in attitude1[var].coords])
            attitude0 = open_level1_attitude(Orbit(orbit - 1), load_pointing=True)
            attitude0 = attitude0.drop('ScanNumber').drop(
                [var for var in attitude0.data_vars if 'mjd' not in attitude0[var].coords])
            self._attitude = xr.concat([attitude0, attitude1], dim='mjd')

        # Correct bouncing attitude quaternions
        self._attitude.Qachieved.values = \
            self._attitude.Qachieved * \
            self._attitude.Qachieved.dot(self._attitude.Qachieved.shift(mjd=1), dims='1ijk').pipe(np.sign).cumprod()

    @property
    def position_eci(self):
        """ Calculate satellite position of a spectrograph / IR Dataset in ECI coordinates. """
        logging.debug('Calculating level1 ECI position.')
        if self._position_eci is None:
            # Reindex mjd to mjdorbitstate to use mjdorbitstate values for position interpolation
            position = self._attitude.reindex(indexers={'mjd': self._attitude.mjdorbitstate},
                                              method='nearest', tolerance=1e-6).position
            self._position_eci = xr.DataArray(position.interp(mjd=self._exposuremjd.values),
                                              coords=[('mjd', self._ds.mjd), ('xyz', ['x', 'y', 'z'])])
        return self._position_eci

    @property
    def position_ecef(self):
        """ Calculate satellite position of a spectrograph / IR Dataset in ECEF coordinates. """
        logging.debug('Calculating level1 ECEF position.')
        if self._position_ecef is None:
            position_ecef = [ECI(*eci).to_geo(mjd) for eci, mjd in zip(self.position_eci, self._exposuremjd)]

            self._position_ecef = xr.DataArray(position_ecef, coords=[('mjd', self._ds.mjd), ('xyz', ['x', 'y', 'z'])])
        return self._position_ecef

    @property
    def position(self):
        """ Calculate satellite position of a spectrograph / IR Dataset.

            Returns the same result as position_eci. """
        msg = 'Level 1 service l1.position will be deprecated in future versions. ' \
              'Use position_ecef or position_eci instead.'
        warnings.warn(msg, FutureWarning, stacklevel=2)
        return self.position_eci

    @property
    def look_eci(self):
        """ Calculate spectrograph / IR look vectors in ECI coordinates. """
        logging.debug('Calculating level1 ECI look vectors.')
        if self._look_eci is None:
            # Load the quaternions
            quaternion = self._attitude.Qachieved
            if 'channel' in self._ds.data_vars:  # it is an ir dataset
                pixel_dict = {1: IR1_PIXEL_PARAMS, 2: IR2_PIXEL_PARAMS, 3: IR3_PIXEL_PARAMS}
                channel = self._ds.channel.item()
                thetas = (self._ds.pixel + 0.5 - pixel_dict[channel].y0) * pixel_dict[channel].dy
                pixel_cfs = [Vector(cos(theta), sin(theta), 0.0) for theta in thetas]
                look = [(UnitQuaternion(*q) * IR_ROTATION_CF).rot_no_z(pixel_cfs)
                        for q in quaternion.interp(mjd=self._exposuremjd.values).values]
                self._look_eci = xr.DataArray(np.array(look),
                                              coords=[('mjd', self._ds.mjd),
                                                      ('pixel', range(128)),
                                                      ('xyz', ['x', 'y', 'z'])])
            else:  # it is a spectrograph dataset
                look = [(UnitQuaternion(*q) * SPECTROGRAPH_ROTATION_CF).rot_unit_x()
                        for q in quaternion.interp(mjd=self._exposuremjd.values).values]
                # Apply the RSAS shift to the spectrograph data if the shift was loaded in the spectrograph data.
                if 'rsas_shift' in self._ds.data_vars:
                    shifted_look = []
                    geo = sk.Geodetic()
                    for (p, l) in zip(self.position_eci.values, look):
                        # A better shift would also slightly change the latitude/longitude, but this is good enough
                        geo.from_tangent_point(p, l)

                        geo.from_lat_lon_alt(geo.latitude, geo.longitude,
                                             geo.altitude - float(self._ds.data_vars['rsas_shift']))
                        new_look = geo.location - p
                        shifted_look.append(new_look / np.linalg.norm(new_look))
                    look = shifted_look
                self._look_eci = xr.DataArray(look, coords=[('mjd', self._ds.mjd), ('xyz', ['x', 'y', 'z'])])
        return self._look_eci

    @property
    def look_ecef(self):
        """ Calculate spectrograph / IR look vectors in ECEF coordinates. """
        logging.debug('Calculating level1 ECEF look vectors.')
        if self._look_ecef is None:
            if 'channel' in self._ds.data_vars:  # it is an ir dataset
                look_ecef = []
                for look_eci, mjd in zip(self.look_eci.values, self._exposuremjd.values):
                    pixel_ecef = [ECI(*pixel_eci).to_geo(mjd) for pixel_eci in look_eci]
                    look_ecef.append(pixel_ecef)
                self._look_ecef = xr.DataArray(np.array(look_ecef),
                                               coords=[('mjd', self._ds.mjd),
                                                       ('pixel', range(128)),
                                                       ('xyz', ['x', 'y', 'z'])])
            else:  # it is a spectrograph dataset
                look_ecef = [ECI(*eci).to_geo(mjd) for eci, mjd in zip(self.look_eci.values, self._exposuremjd.values)]
                self._look_ecef = xr.DataArray(look_ecef, coords=[('mjd', self._ds.mjd), ('xyz', ['x', 'y', 'z'])])
        return self._look_ecef

    @property
    def look(self):
        """ Calculate spectrograph / IR look vectors.

            Returns the same result as look_eci. """
        msg = 'Level 1 service l1.look will be deprecated in future versions. ' \
              'Use look_ecef or look_eci instead.'
        warnings.warn(msg, FutureWarning, stacklevel=2)
        return self.look_eci

    def tangent_point(self):
        """ Calculate spectrograph / IR tangent information (lat, lon, altitude). """
        logging.debug('Calculating level1 tangent points.')

        latitude = []
        longitude = []
        altitude = []
        geo = sk.Geodetic()
        if self.look_ecef.ndim == 3:  # it is an ir dataset
            for (mjd, pos, look) in zip(self._exposuremjd.values, self.position_ecef.values, self.look_ecef.values):
                for pixellook in look:
                    geo.from_tangent_point(Vector(*pos), Vector(*pixellook))
                    # Check that the tangent altitude is above the center of the Earth.
                    if geo.altitude > -6000000.:
                        latitude.append(geo.latitude)
                        longitude.append(geo.longitude)
                        altitude.append(geo.altitude)
                    else:  # return NaN
                        latitude.append(np.nan)
                        longitude.append(np.nan)
                        altitude.append(np.nan)
            self._latitude = xr.DataArray(np.reshape(latitude, (-1, 128)),
                                          coords=[('mjd', self._ds.mjd), ('pixel', range(128))])
            self._longitude = xr.DataArray(np.reshape(longitude, (-1, 128)),
                                           coords=[('mjd', self._ds.mjd), ('pixel', range(128))])
            self._altitude = xr.DataArray(np.reshape(altitude, (-1, 128)),
                                          coords=[('mjd', self._ds.mjd), ('pixel', range(128))])
        else:  # it is a spectrograph dataset
            for (mjd, pos, look) in zip(self._exposuremjd.values, self.position_ecef.values, self.look_ecef.values):
                geo.from_tangent_point(Vector(*pos), Vector(*look))
                latitude.append(geo.latitude)
                longitude.append(geo.longitude)
                altitude.append(geo.altitude)
            self._latitude = xr.DataArray(latitude, coords=[('mjd', self._ds.mjd)])
            self._longitude = xr.DataArray(longitude, coords=[('mjd', self._ds.mjd)])
            self._altitude = xr.DataArray(altitude, coords=[('mjd', self._ds.mjd)])

    @property
    def latitude(self):
        if self._latitude is None:
            self.tangent_point()
        return self._latitude

    @property
    def longitude(self):
        if self._longitude is None:
            self.tangent_point()
        return self._longitude

    @property
    def altitude(self):
        if self._altitude is None:
            self.tangent_point()
        return self._altitude

    @property
    def sza(self):
        """ Calculate spectrograph / IR solar zenith angle. """
        logging.debug('Calculating level1 tangent points.')
        if self._sza is None:
            sza = []
            if self.look_ecef.ndim == 3:  # it is an ir dataset
                for (mjd, lat, lon, height) in zip(self._exposuremjd.values, self.latitude.values,
                                                   self.longitude.values, self.altitude.values):
                    if np.isnan(height).any():
                        sza.append([np.nan, np.nan])
                    else:
                        time = Time(mjd, format='mjd')
                        sun = get_sun(time)
                        location_px0 = EarthLocation(lat=lat[0], lon=lon[0], height=height[0])
                        location_px127 = EarthLocation(lat=lat[127], lon=lon[127], height=height[127])
                        sza_px0 = (90.0 - sun.transform_to(AltAz(obstime=time, location=location_px0)).alt.value)
                        sza_px127 = (90.0 - sun.transform_to(AltAz(obstime=time, location=location_px127)).alt.value)
                        sza.append([sza_px0, sza_px127])
                sza = xr.DataArray(sza, coords=[('mjd', self._ds.mjd), ('pixel', [0, 127])])
                self._sza = sza.interp(pixel=range(128))
            else:  # it is a spectrograph dataset
                for (mjd, lat, lon, height) in zip(self._exposuremjd.values, self.latitude.values,
                                                   self.longitude.values, self.altitude.values):
                    location = EarthLocation(lat=lat, lon=lon, height=height)
                    time = Time(mjd, format='mjd')
                    sun = get_sun(time).transform_to(AltAz(obstime=time, location=location))
                    sza.append(90.0 - sun.alt.value)
                self._sza = xr.DataArray(sza, coords=[('mjd', self._ds.mjd)])
        return self._sza


def get_scannos_in_orbit_below_sza(orbit: int, sza: float, target_altitude: float = 25000):
    """ For a given orbit return an iterator of the scans whose tangent point is above a configurable horizon.
        Useful for retrieval code.

        :param Orbit orbit: The orbit class handling the data retrieval, specified by orbit number
        :param float sza: Reference solar zenith angle
        :param target_altitude: Altitude to take solar zenith angle reference from, default is 25000m
    """
    try:
        attitude = open_level1_attitude(orbit=orbit, load_pointing=True)
    except LookupError:  # No valid orbits
        return

    for scanno in attitude.ScanNumber.values:
        logging.debug(f'Checking scan {scanno} to see if it is below a sza of {sza}')
        try:
            spectrograph = open_level1_spectrograph(scanno=scanno)
        except LookupError:  # No data for scan
            logging.debug(f'Scan {scanno} has no spectrograph data.')
            continue

        spectrograph['altitude'] = spectrograph.l1.altitude
        spectrograph['sza'] = spectrograph.l1.sza
        spectrograph = spectrograph.dropna(dim='mjd', subset=['altitude']).sortby('altitude')

        sza_at_target = spectrograph.swap_dims({'mjd': 'altitude'}).sza.sel(altitude=target_altitude, method='nearest')

        if sza_at_target < sza:
            logging.info(f'SZA for scan {scanno} is below {sza}, yielding scan.')
            yield scanno
        else:
            logging.info(f'SZA for scan {scanno} is not below {sza}, skipping scan.')
